from cogdl.models.nn.correct_smooth import CorrectSmooth
from .dgi_mw import DGIModelWrapper
from .gcnmix_mw import GCNMixModelWrapper
from .grace_mw import GRACEModelWrapper
from .grand_mw import GrandModelWrapper
from .graphsage_mw import GraphSAGEModelWrapper
from .unsup_graphsage_mw import UnsupGraphSAGEModelWrapper
from .m3s_mw import M3SModelWrapper
from .mvgrl_mw import MVGRLModelWrapper
from .network_embedding_mw import NetworkEmbeddingModelWrapper
from .node_classification_mw import NodeClfModelWrapper
from .pprgo_mw import PPRGoModelWrapper
from .sagn_mw import SAGNModelWrapper
from .self_auxiliary_mw import SelfAuxiliaryModelWrapper
from .correct_smooth_mw import CorrectSmoothModelWrapper
