class Logger:
    def __init__(self, log_path):
        self.log_path = log_path

    def start(self):
        pass

    def note(self, metrics, step=None):
        pass

    def finish(self):
        pass
